//
//  GameScene.swift
//  SchoolhouseSkateboarder
//
//  Created by MATTHEW MCCARTHY on 4/5/17.
//  Copyright © 2017 iOS Kids. All rights reserved.
//

import SpriteKit

/// Deze struct bevat verschillende fysieke categorieen waarmee we kunnen definieren welke objecttypes botsen of contact met elkaar hebben
struct PhysicsCategory {
    static let skater: UInt32 = 0x1 << 0
    static let steen: UInt32 = 0x1 << 1
    static let gem: UInt32 = 0x1 << 2
}

class GameScene: SKScene, SKPhysicsContactDelegate {
    
    // Een array die alle huidige stoeptegels bevat
    var stenen = [SKSpriteNode]()
    
    // Het formaat van de stoeptegel graphics is gebruikt
    var steenSize = CGSize.zero
    
    // Instelling voor hoe snel het spel naar rechts scrolt
    // Dit kan toenemen als de gebruiker vordert in het spel
    var scrollSnelheid: CGFloat = 5.0
    let startScrollsnelheid: CGFloat = 5.0
    
    // Een constante voor zwaartekracht of hoe snel objecten op de aarde vallen
    let zwaartekrachtSnelheid: CGFloat = 1.5
    
    // Het tijdstip van de laatste update
    var laatsteUpdateTijdstip: TimeInterval?
    
    // De held van het spel, de skater, is hier gemaakt
    let skater = Skater(imageNamed: "skater")
    
    override func didMove(to view: SKView) {
        
        physicsWorld.gravity = CGVector(dx: 0.0, dy: -6.0)
        physicsWorld.contactDelegate = self
        
        anchorPoint = CGPoint.zero
        
        let achtergrond = SKSpriteNode(imageNamed: "achtergrond")
        let xMid = frame.midX
        let yMid = frame.midY
        achtergrond.position = CGPoint(x: xMid, y: yMid)
        addChild(achtergrond)
        
        // De skater instellen en toevoegen aan de scene
        skater.setupPhysicsBody()
        addChild(skater)
        
        // Voeg een tik gesture recognizer toe om te weten wanneer de gebruiker op het scherm heeft getikt
        let tikMethode = #selector(GameScene.behandelTik(tikGesture:))
        let tikGesture = UITapGestureRecognizer(target: self, action: tikMethode)
        view.addGestureRecognizer(tikGesture)
        
        startGame()
    }
    
    func resetSkater() {
        
        // Stel de startpositie, zPosition en minimumY van de skater in
        let skaterX = frame.midX / 2.0
        let skaterY = skater.frame.height / 2.0 + 64.0
        skater.position = CGPoint(x: skaterX, y: skaterY)
        skater.zPosition = 10
        skater.minimumY = skaterY
        
        skater.zRotation = 0.0
        skater.physicsBody?.velocity = CGVector(dx: 0.0, dy: 0.0)
        skater.physicsBody?.angularVelocity = 0.0
    }
    
    func startGame() {
        
        // Als een nieuw spel is gestart, reset naar de startinstellingen
        
        resetSkater()
        
        scrollSnelheid = startScrollsnelheid
        laatsteUpdateTijdstip = nil
        
        for steen in stenen {
            steen.removeFromParent()
        }
        
        stenen.removeAll(keepingCapacity: true)
    }
    
    func gameOver() {
        
        startGame()
    }
    
    func spawnSteen(opPositie position: CGPoint) -> SKSpriteNode {
        
        // Maak een steen sprite en voeg het aan de scene toe
        let steen = SKSpriteNode(imageNamed: "stoep")
        steen.position = position
        steen.zPosition = 8
        addChild(steen)
        
        // Update de steenSize door de echte steen size
        steenSize = steen.size
        
        // Voeg de nieuwe steen toe aan het array van stenen
        stenen.append(steen)
        
        // Stel het fysieke lichaam van de steen in
        let center = steen.centerRect.origin
        steen.physicsBody = SKPhysicsBody(rectangleOf: steen.size, center: center)
        steen.physicsBody?.affectedByGravity = false
        
        steen.physicsBody?.categoryBitMask = PhysicsCategory.steen
        steen.physicsBody?.collisionBitMask = 0
        
        // Retourneer deze steen naar de aanroeper
        return steen
    }
    
    func updateStenen(metScrollhoeveelheid huidigeScrollhoeveelheid: CGFloat) {
        
        // Houd de grootste x-positie van alle huidige stenen bij
        var versteRechtersteenX: CGFloat = 0.0
        
        for steen in stenen {
            
            let nieuweX = steen.position.x - huidigeScrollhoeveelheid
            
            // Als een steen te ver naar links is verschoven (uit het scherm), verwijder hem dan
            if nieuweX < -steenSize.width {
                
                steen.removeFromParent()
                
                if let steenIndex = stenen.index(of: steen) {
                    stenen.remove(at: steenIndex)
                }
                
            } else {
                
                // Update de positie van een steen die nog op het scherm staat
                steen.position = CGPoint(x: nieuweX, y: steen.position.y)
                
                // Update de meest rechtse positietracker
                if steen.position.x > versteRechtersteenX {
                    versteRechtersteenX = steen.position.x
                }
            }
        }
        
        // Een while loop om ervoor te zorgen dat ons scherm altijd stenen bevat
        while versteRechtersteenX < frame.width {
            
            var steenX = versteRechtersteenX + steenSize.width + 1.0
            let steenY = steenSize.height / 2.0
            
            // Plaats op willekeurige plekken gaten in de stoep, zodat de speler er overheen moet springen
            let willekeurigNummer = arc4random_uniform(99)
            
            if willekeurigNummer < 5 {
                
                // Er is 5 procent kans dat er een gat gemaakt wordt tussen de stenen
                let gat = 20.0 * scrollSnelheid
                steenX += gat
            }
            
            // Laat een nieuwe steen verschijnen en update de meest rechtse steen
            let nieuweSteen = spawnSteen(opPositie: CGPoint(x: steenX, y: steenY))
            versteRechtersteenX = nieuweSteen.position.x
        }
    }
    
    func updateSkater() {
        
        // Bepaal of de skater zich op de grond bevindt
        if let snelheidY = skater.physicsBody?.velocity.dy {
            
            if snelheidY < -100.0 || snelheidY > 100.0 {
                skater.isOpGrond = false
            }
        }
        
        // Controleer of het spel beeindigd moet worden
        let isOffScreen = skater.position.y < 0.0 || skater.position.x < 0.0
        
        let maxRotatie = CGFloat(GLKMathDegreesToRadians(85.0))
        let isOvergeslagen = skater.zRotation > maxRotatie || skater.zRotation < -maxRotatie
        
        if isOffScreen || isOvergeslagen {
            gameOver()
        }
    }
    
    override func update(_ currentTime: TimeInterval) {
        
        // Bepaal de verstreken tijd sinds de laatste update is aangeroepen
        var verstrekenTijd: TimeInterval = 0.0
        
        if let laatsteTijdstip = laatsteUpdateTijdstip {
            verstrekenTijd = currentTime - laatsteTijdstip
        }
        
        laatsteUpdateTijdstip = currentTime
        
        let verwachteVerstrekenTijd: TimeInterval = 1.0 / 60.0
        
        // Hier berekenen we hoe ver alles moet bewegen in deze update
        let scrollAanpassing = CGFloat(verstrekenTijd / verwachteVerstrekenTijd)
        let huidigeScrollhoeveelheid = scrollSnelheid * scrollAanpassing
        
        updateStenen(metScrollhoeveelheid: huidigeScrollhoeveelheid)
        
        updateSkater()
    }
    
    func behandelTik(tikGesture: UITapGestureRecognizer) {
        
        // Laat de skater springen als de speler op het scherm tikt wanneer de skater zich op de grond bevindt
        if skater.isOpGrond {
            
            skater.physicsBody?.applyImpulse(CGVector(dx: 0.0, dy: 260.0))
        }
    }
    
    // MARK:- SKPhysicsContactDelegate Methods
    func didBegin(_ contact: SKPhysicsContact) {
        
        // Controleer of er contact is tussen de skater en een steen
        if contact.bodyA.categoryBitMask == PhysicsCategory.skater && contact.bodyB.categoryBitMask == PhysicsCategory.steen {
            
            skater.isOpGrond = true
        }
    }
}
