//
//  GameScene.swift
//  SchoolhouseSkateboarder
//
//  Created by MATTHEW MCCARTHY on 4/5/17.
//  Copyright © 2017 iOS Kids. All rights reserved.
//

import SpriteKit

/// Deze struct bevat verschillende fysieke categorieen waarmee we kunnen definieren welke objecttypes botsen of contact met elkaar hebben
struct PhysicsCategory {
    static let skater: UInt32 = 0x1 << 0
    static let steen: UInt32 = 0x1 << 1
    static let gem: UInt32 = 0x1 << 2
}

class GameScene: SKScene, SKPhysicsContactDelegate {
    
    // Enum voor y-positie spawn punten voor stenen
    // Stenen op de grond zijn laag en de platformstenen zijn hoog
    enum SteenLevel: CGFloat {
        
        case low = 0.0
        case high = 100.0
    }
    
    
    // MARK:- Class Properties
    
    // Een array die alle huidige stoeptegels bevat
    var stenen = [SKSpriteNode]()
    
    // Een array dat alle huidige gems bevat
    var gems = [SKSpriteNode]()
    
    // Het formaat van de stoeptegel graphics is gebruikt
    var steenSize = CGSize.zero
    
    // Het huidige steenlevel bepaalt de y-positie van nieuwe stenen
    var steenLevel = SteenLevel.low
    
    // Instelling voor hoe snel het spel naar rechts scrolt
    // Dit kan toenemen als de gebruiker vordert in het spel
    var scrollSnelheid: CGFloat = 5.0
    let startScrollsnelheid: CGFloat = 5.0
    
    // Een constante voor zwaartekracht of hoe snel objecten op de aarde vallen
    let zwaartekrachtSnelheid: CGFloat = 1.5
    
    // Eigenschappen voor het bijhouden van de score
    var score: Int = 0
    var highScore: Int = 0
    var laatsteScoreUpdateTijdstip: TimeInterval = 0.0
    
    // Het tijdstip van de laatste update
    var laatsteUpdateTijdstip: TimeInterval?
    
    // De held van het spel, de skater, is hier gemaakt
    let skater = Skater(imageNamed: "skater")
    
    
    // MARK:- Setup and Lifecycle Methods
    
    override func didMove(to view: SKView) {
        
        physicsWorld.gravity = CGVector(dx: 0.0, dy: -6.0)
        physicsWorld.contactDelegate = self
        
        anchorPoint = CGPoint.zero
        
        let achtergrond = SKSpriteNode(imageNamed: "achtergrond")
        let xMid = frame.midX
        let yMid = frame.midY
        achtergrond.position = CGPoint(x: xMid, y: yMid)
        addChild(achtergrond)
        
        setupLabels()
        
        // De skater instellen en toevoegen aan de scene
        skater.setupPhysicsBody()
        addChild(skater)
        
        // Voeg een tik gesture recognizer toe om te weten wanneer de gebruiker op het scherm heeft getikt
        let tikMethode = #selector(GameScene.behandelTik(tikGesture:))
        let tikGesture = UITapGestureRecognizer(target: self, action: tikMethode)
        view.addGestureRecognizer(tikGesture)
        
        startGame()
    }
    
    func resetSkater() {
        
        // Stel de startpositie, zPosition en minimumY van de skater in
        let skaterX = frame.midX / 2.0
        let skaterY = skater.frame.height / 2.0 + 64.0
        skater.position = CGPoint(x: skaterX, y: skaterY)
        skater.zPosition = 10
        skater.minimumY = skaterY
        
        skater.zRotation = 0.0
        skater.physicsBody?.velocity = CGVector(dx: 0.0, dy: 0.0)
        skater.physicsBody?.angularVelocity = 0.0
    }
    
    func setupLabels() {
        
        // Label met "score" in de linkerbovenhoek
        
        let scoreTextLabel: SKLabelNode = SKLabelNode(text: "score")
        scoreTextLabel.position = CGPoint(x: 14.0, y: frame.size.height - 20.0)
        scoreTextLabel.horizontalAlignmentMode = .left
        scoreTextLabel.fontName = "Courier-Bold"
        scoreTextLabel.fontSize = 14.0
        scoreTextLabel.zPosition = 20
        addChild(scoreTextLabel)
        
        // Label dat de huidige score van de speler toont
        
        let scoreLabel: SKLabelNode = SKLabelNode(text: "0")
        scoreLabel.position = CGPoint(x: 14.0, y: frame.size.height - 40.0)
        scoreLabel.horizontalAlignmentMode = .left
        scoreLabel.fontName = "Courier-Bold"
        scoreLabel.fontSize = 18.0
        scoreLabel.name = "scoreLabel"
        scoreLabel.zPosition = 20
        addChild(scoreLabel)
        
        // Label dat "high score" in de rechterbovenhoek toont
        
        let highScoreTextLabel: SKLabelNode = SKLabelNode(text: "high score")
        highScoreTextLabel.position = CGPoint(x: frame.size.width - 14.0, y: frame.size.height - 20.0)
        highScoreTextLabel.horizontalAlignmentMode = .right
        highScoreTextLabel.fontName = "Courier-Bold"
        highScoreTextLabel.fontSize = 14.0
        highScoreTextLabel.zPosition = 20
        addChild(highScoreTextLabel)
        
        // Label dat de huidige hoogste score van de speler toont
        
        let highScoreLabel: SKLabelNode = SKLabelNode(text: "0")
        highScoreLabel.position = CGPoint(x: frame.size.width - 14.0, y: frame.size.height - 40.0)
        highScoreLabel.horizontalAlignmentMode = .right
        highScoreLabel.fontName = "Courier-Bold"
        highScoreLabel.fontSize = 18.0
        highScoreLabel.name = "highScoreLabel"
        highScoreLabel.zPosition = 20
        addChild(highScoreLabel)
    }
    
    func updateScoreLabelText() {
        
        if let scoreLabel = childNode(withName: "scoreLabel") as? SKLabelNode {
            scoreLabel.text = String(format: "%04d", score)
        }
    }
    
    func updateHighScoreLabelText() {
        
        if let highScoreLabel = childNode(withName: "highScoreLabel") as? SKLabelNode {
            highScoreLabel.text = String(format: "%04d", highScore)
        }
    }
    
    func startGame() {
        
        // Als een nieuw spel is gestart, reset naar de startinstellingen
        
        resetSkater()
        
        score = 0
        
        scrollSnelheid = startScrollsnelheid
        steenLevel = .low
        laatsteUpdateTijdstip = nil
        
        for steen in stenen {
            steen.removeFromParent()
        }
        
        stenen.removeAll(keepingCapacity: true)
        
        for gem in gems {
            verwijderGem(gem)
        }
    }
    
    func gameOver() {
        
        // Bekijk of de speler een nieuwe high score heeft als het spel af is
        
        if score > highScore {
            highScore = score
            
            updateHighScoreLabelText()
        }
        
        startGame()
    }
    
    
    // MARK:- Spawn and Remove Methods
    
    func spawnSteen(opPositie position: CGPoint) -> SKSpriteNode {
        
        // Maak een steen sprite en voeg het aan de scene toe
        let steen = SKSpriteNode(imageNamed: "stoep")
        steen.position = position
        steen.zPosition = 8
        addChild(steen)
        
        // Update de steenSize door de echte steen size
        steenSize = steen.size
        
        // Voeg de nieuwe steen toe aan het array van stenen
        stenen.append(steen)
        
        // Stel het fysieke lichaam van de steen in
        let center = steen.centerRect.origin
        steen.physicsBody = SKPhysicsBody(rectangleOf: steen.size, center: center)
        steen.physicsBody?.affectedByGravity = false
        
        steen.physicsBody?.categoryBitMask = PhysicsCategory.steen
        steen.physicsBody?.collisionBitMask = 0
        
        // Retourneer deze steen naar de aanroeper
        return steen
    }
    
    func spawnGem(opPositie position: CGPoint) {
        
        // Maak een gem sprite en voeg het toe aan de scene
        let gem = SKSpriteNode(imageNamed: "gem")
        gem.position = position
        gem.zPosition = 9
        addChild(gem)
        
        gem.physicsBody = SKPhysicsBody(rectangleOf: gem.size, center: gem.centerRect.origin)
        gem.physicsBody?.categoryBitMask = PhysicsCategory.gem
        gem.physicsBody?.affectedByGravity = false
        
        // Voeg een nieuwe gem toe aan de array van gems
        gems.append(gem)
    }
    
    func verwijderGem(_ gem: SKSpriteNode) {
        
        gem.removeFromParent()
        
        if let gemIndex = gems.index(of: gem) {
            gems.remove(at: gemIndex)
        }
    }
    
    
    // MARK:- Update Methods
    
    func updateStenen(metScrollhoeveelheid huidigeScrollhoeveelheid: CGFloat) {
        
        // Houd de grootste x-positie van alle huidige stenen bij
        var versteRechtersteenX: CGFloat = 0.0
        
        for steen in stenen {
            
            let nieuweX = steen.position.x - huidigeScrollhoeveelheid
            
            // Als een steen te ver naar links is verschoven (uit het scherm), verwijder hem dan
            if nieuweX < -steenSize.width {
                
                steen.removeFromParent()
                
                if let steenIndex = stenen.index(of: steen) {
                    stenen.remove(at: steenIndex)
                }
                
            } else {
                
                // Update de positie van een steen die nog op het scherm staat
                steen.position = CGPoint(x: nieuweX, y: steen.position.y)
                
                // Update de meest rechtse positietracker
                if steen.position.x > versteRechtersteenX {
                    versteRechtersteenX = steen.position.x
                }
            }
        }
        
        // Een while loop om ervoor te zorgen dat ons scherm altijd stenen bevat
        while versteRechtersteenX < frame.width {
            
            var steenX = versteRechtersteenX + steenSize.width + 1.0
            let steenY = (steenSize.height / 2.0) + steenLevel.rawValue
            
            // Plaats op willekeurige plekken gaten in de stoep, zodat de speler er overheen moet springen
            let willekeurigNummer = arc4random_uniform(99)
            
            if willekeurigNummer < 2 && score > 10 {
                
                // Er is 2 procent kans dat er een gat gemaakt wordt tussen de stenen nadat de speler de score van 10 heeft bereikt
                let gat = 20.0 * scrollSnelheid
                steenX += gat
                
                // Voeg bij ieder gat een gem toe
                let willekeurigeGemYHoeveelheid = CGFloat(arc4random_uniform(150))
                let nieuweGemY = steenY + skater.size.height + willekeurigeGemYHoeveelheid
                let nieuweGemX = steenX - gat / 2.0
                
                spawnGem(opPositie: CGPoint(x: nieuweGemX, y: nieuweGemY))
            }
            else if willekeurigNummer < 4 && score > 20 {
                
                // Er is een kans van 2 procent dat het steenlevel verandert nadat de speler de score van 20 heeft bereikt
                if steenLevel == .high {
                    steenLevel = .low
                }
                else if steenLevel == .low {
                    steenLevel = .high
                }
            }
            
            // Laat een nieuwe steen verschijnen en update de meest rechtse steen
            let nieuweSteen = spawnSteen(opPositie: CGPoint(x: steenX, y: steenY))
            versteRechtersteenX = nieuweSteen.position.x
        }
    }
    
    func updateGems(metScrollhoeveelheid huidigeScrollhoeveelheid: CGFloat) {
        
        for gem in gems {
            
            // Update iedere gem-positie
            let dezeGemX = gem.position.x - huidigeScrollhoeveelheid
            gem.position = CGPoint(x: dezeGemX, y: gem.position.y)

            // Verwijder alle gems die van het scherm zijn verdwenen
            if gem.position.x < 0.0 {
                
                verwijderGem(gem)
            }
        }
    }
    
    func updateSkater() {
        
        // Bepaal of de skater zich op de grond bevindt
        if let snelheidY = skater.physicsBody?.velocity.dy {
            
            if snelheidY < -100.0 || snelheidY > 100.0 {
                skater.isOpGrond = false
            }
        }
        
        // Controleer of het spel beeindigd moet worden
        let isOffScreen = skater.position.y < 0.0 || skater.position.x < 0.0
        
        let maxRotatie = CGFloat(GLKMathDegreesToRadians(85.0))
        let isOvergeslagen = skater.zRotation > maxRotatie || skater.zRotation < -maxRotatie
        
        if isOffScreen || isOvergeslagen {
            gameOver()
        }
    }
    
    func updateScore(withCurrentTime currentTime: TimeInterval) {
        
        // De score van de speler wordt hoger naarmate ze langer overleven
        // Update de score alleen iedere seconde
        
        let verstrekenTijd = currentTime - laatsteScoreUpdateTijdstip
        
        if verstrekenTijd > 1.0 {
            
            // Verhoog de score
            score += Int(scrollSnelheid)
            
            // Reset het laasteScoreUpdateTijdstip naar de huidige tijd
            laatsteScoreUpdateTijdstip = currentTime
            
            updateScoreLabelText()
        }
    }
    
    
    // MARK:- Main Game Loop Method
    
    override func update(_ currentTime: TimeInterval) {
        
        // Verhoog langzaamaan de scrollSnelheid terwijl het spel vordert
        scrollSnelheid += 0.01
        
        // Bepaal de verstreken tijd sinds de laatste update is aangeroepen
        var verstrekenTijd: TimeInterval = 0.0
        
        if let laatsteTijdstip = laatsteUpdateTijdstip {
            verstrekenTijd = currentTime - laatsteTijdstip
        }
        
        laatsteUpdateTijdstip = currentTime
        
        let verwachteVerstrekenTijd: TimeInterval = 1.0 / 60.0
        
        // Hier berekenen we hoe ver alles moet bewegen in deze update
        let scrollAanpassing = CGFloat(verstrekenTijd / verwachteVerstrekenTijd)
        let huidigeScrollhoeveelheid = scrollSnelheid * scrollAanpassing
        
        updateStenen(metScrollhoeveelheid: huidigeScrollhoeveelheid)
        updateSkater()
        updateGems(metScrollhoeveelheid: huidigeScrollhoeveelheid)
        updateScore(withCurrentTime: currentTime)
    }
    
    
    // MARK:- Touch Handling Methods
    
    func behandelTik(tikGesture: UITapGestureRecognizer) {
        
        // Laat de skater springen als de speler op het scherm tikt wanneer de skater zich op de grond bevindt
        if skater.isOpGrond {
            
            skater.physicsBody?.applyImpulse(CGVector(dx: 0.0, dy: 260.0))
        }
    }
    
    
    // MARK:- SKPhysicsContactDelegate Methods
    
    func didBegin(_ contact: SKPhysicsContact) {
        
        // Controleer of er contact is tussen de skater en een steen
        if contact.bodyA.categoryBitMask == PhysicsCategory.skater && contact.bodyB.categoryBitMask == PhysicsCategory.steen {
            
            skater.isOpGrond = true
        }
        else if contact.bodyA.categoryBitMask == PhysicsCategory.skater && contact.bodyB.categoryBitMask == PhysicsCategory.gem {
            
            // Skater heeft een gem geraakt, dus verwijder het
            if let gem = contact.bodyB.node as? SKSpriteNode {
                
                verwijderGem(gem)
                
                // Geef de speler 50 punten voor het verzamelen van een gem
                score += 50
                updateScoreLabelText()
            }
        }
    }
}
